@@ -11,10 +11,13 @@
 import abc
 import contextlib
 import math
+import sys
 import warnings
 from collections.abc import Iterable
+from functools import cached_property
 from random import Random
 from sys import float_info
+from types import ModuleType
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -45,7 +48,11 @@ from hypothesis.internal.conjecture.utils import (
     Sampler,
     many,
 )
-from hypothesis.internal.constants_ast import local_constants
+from hypothesis.internal.constants_ast import (
+    Constants,
+    constants_from_module,
+    is_local_module_file,
+)
 from hypothesis.internal.floats import (
     SIGNALING_NAN,
     float_to_int,
@@ -59,7 +66,7 @@ if TYPE_CHECKING:
     from typing import TypeAlias
 
     from hypothesis.internal.conjecture.data import ConjectureData
-    from hypothesis.internal.constants_ast import ConstantsT, ConstantT
+    from hypothesis.internal.constants_ast import ConstantT
 
 T = TypeVar("T")
 _Lifetime: "TypeAlias" = Literal["test_case", "test_function"]
@@ -80,7 +87,6 @@ AVAILABLE_PROVIDERS = {
     "hypothesis": "hypothesis.internal.conjecture.providers.HypothesisProvider",
     "hypothesis-urandom": "hypothesis.internal.conjecture.providers.URandomProvider",
 }
-FLOAT_INIT_LOGIC_CACHE = LRUCache(4096)
 # cache the choice_permitted constants for a particular set of constraints.
 CONSTANTS_CACHE = LRUCache(1024)
 
@@ -189,29 +195,76 @@ _constant_strings = {
 
 # we don't actually care what order the constants are sorted in, just that the
 # ordering is deterministic.
-GLOBAL_CONSTANTS: "ConstantsT" = {
-    "float": SortedSet(_constant_floats, key=float_to_int),
-    "string": SortedSet(_constant_strings),
-    "integer": SortedSet(),
-    "bytes": SortedSet(),
-}
+GLOBAL_CONSTANTS = Constants(
+    integers=SortedSet(),
+    floats=SortedSet(_constant_floats, key=float_to_int),
+    bytes=SortedSet(),
+    strings=SortedSet(_constant_strings),
+)
 
 
-_local_constants_hash: Optional[int] = None
+_local_constants = Constants(
+    integers=SortedSet(),
+    floats=SortedSet(key=float_to_int),
+    bytes=SortedSet(),
+    strings=SortedSet(),
+)
+# modules that we've already seen and processed for local constants. These are
+# are all modules, not necessarily local ones. This lets us quickly see which
+# modules are new without an expensive path.resolve() or is_local_module_file
+# cache lookup.
+_seen_modules: set[ModuleType] = set()
+_sys_modules_len: Optional[int] = None
 
 
-def _get_local_constants():
-    global _local_constants_hash
+def _get_local_constants() -> Constants:
+    global _sys_modules_len, _local_constants
 
-    constants = local_constants()
-    constants_hash = hash(tuple((k, tuple(v)) for k, v in constants.items()))
-    # if we've added new constants since the last time we checked, invalidate
-    # the cache.
-    if constants_hash != _local_constants_hash:
+    if sys.platform == "emscripten":  # pragma: no cover
+        # pyodide builds bundle the stdlib in a nonstandard location, like
+        # `/lib/python312.zip/heapq.py`. To avoid identifying the entirety of
+        # the stdlib as local code and slowing down on emscripten, instead return
+        # that nothing is local.
+        #
+        # pyodide may provide some way to distinguish stdlib/third-party/local
+        # code. I haven't looked into it. If they do, we should correctly implement
+        # ModuleLocation for pyodide instead of this.
+        return _local_constants
+
+    count_constants = len(_local_constants)
+    # We call this function once per HypothesisProvider instance, i.e. once per
+    # input, so it needs to be performant. The logic here is more complicated
+    # than necessary because of this.
+    #
+    # First, we check whether there are any new modules with a very cheap length
+    # check. This check can be fooled if a module is added while another module is
+    # removed, but the more correct check against tuple(sys.modules.keys()) is
+    # substantially more expensive. Such a new module would eventually be discovered
+    # if / when the length changes again in the future.
+    #
+    # If the length has changed, we find just modules we haven't seen before. Of
+    # those, we find the ones which correspond to local modules, and extract their
+    # constants.
+
+    # careful: store sys.modules length when we first check to avoid race conditions
+    # with other threads loading a module before we set _sys_modules_len.
+    if (sys_modules_len := len(sys.modules)) != _sys_modules_len:
+        new_modules = set(sys.modules.values()) - _seen_modules
+        for module in new_modules:
+            if getattr(module, "__file__", None) is not None and is_local_module_file(
+                module.__file__
+            ):
+                _local_constants |= constants_from_module(module)
+        _seen_modules.update(new_modules)
+        _sys_modules_len = sys_modules_len
+
+    # if we add any new constant, invalidate the constant cache for permitted values.
+    # A more efficient approach would be invalidating just the keys with this
+    # choice_type.
+    if len(_local_constants) > count_constants:
         CONSTANTS_CACHE.cache.clear()
-        _local_constants_hash = constants_hash
 
-    return constants
+    return _local_constants
 
 
 class _BackendInfoMsg(TypedDict):
@@ -371,9 +424,13 @@ class HypothesisProvider(PrimitiveProvider):
 
     def __init__(self, conjecturedata: Optional["ConjectureData"], /):
         super().__init__(conjecturedata)
-        self.local_constants = _get_local_constants()
         self._random = None if self._cd is None else self._cd._random
 
+    @cached_property
+    def _local_constants(self):
+        # defer computation of local constants until/if we need it
+        return _get_local_constants()
+
     def _maybe_draw_constant(
         self,
         choice_type: ChoiceTypeT,
@@ -382,6 +439,7 @@ class HypothesisProvider(PrimitiveProvider):
         p: float = 0.05,
     ) -> Optional["ConstantT"]:
         assert self._random is not None
+        assert self._local_constants is not None
         assert choice_type != "boolean"
 
         # check whether we even want a constant before spending time computing
@@ -394,12 +452,12 @@ class HypothesisProvider(PrimitiveProvider):
             CONSTANTS_CACHE[key] = (
                 tuple(
                     choice
-                    for choice in GLOBAL_CONSTANTS[choice_type]
+                    for choice in GLOBAL_CONSTANTS.set_for_type(choice_type)
                     if choice_permitted(choice, constraints)
                 ),
                 tuple(
                     choice
-                    for choice in self.local_constants[choice_type]
+                    for choice in self._local_constants.set_for_type(choice_type)
                     if choice_permitted(choice, constraints)
                 ),
             )
